import datetime
import json
import os
from email.message import EmailMessage
from typing import Any, Dict, List, Optional
from unittest import mock

import pytest

from b4.review import checks


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _make_msg(subject: str = 'test patch', msgid: str = 'abc@example.com',
              body: str = 'dummy') -> EmailMessage:
    """Create a minimal EmailMessage for testing."""
    msg = EmailMessage()
    msg['Subject'] = subject
    msg['Message-Id'] = f'<{msgid}>'
    msg.set_content(body)
    return msg


# ---------------------------------------------------------------------------
# SQLite cache: store / retrieve / delete / cleanup
# ---------------------------------------------------------------------------

class TestCacheDb:
    """Tests for the CI check cache database."""

    def test_get_db_creates_schema(self, tmp_path: pytest.TempPathFactory) -> None:
        conn = checks.get_db()
        cursor = conn.execute(
            "SELECT name FROM sqlite_master WHERE type='table' ORDER BY name"
        )
        tables = [row[0] for row in cursor.fetchall()]
        assert 'check_results' in tables
        assert 'schema_version' in tables
        conn.close()

    def test_store_and_retrieve(self, tmp_path: pytest.TempPathFactory) -> None:
        conn = checks.get_db()
        results = [
            {'tool': 'lint', 'status': 'pass', 'summary': 'ok',
             'url': '', 'details': ''},
            {'tool': 'build', 'status': 'fail', 'summary': 'broken',
             'url': 'https://ci.example.com', 'details': 'error on line 5'},
        ]
        checks.store_results(conn, 'msg1@example', results)
        cached = checks.get_cached_results(conn, ['msg1@example'])
        assert 'msg1@example' in cached
        assert len(cached['msg1@example']) == 2
        tools = {r['tool'] for r in cached['msg1@example']}
        assert tools == {'lint', 'build'}
        conn.close()

    def test_retrieve_empty(self, tmp_path: pytest.TempPathFactory) -> None:
        conn = checks.get_db()
        cached = checks.get_cached_results(conn, ['nonexistent@example'])
        assert cached == {}
        conn.close()

    def test_retrieve_empty_list(self, tmp_path: pytest.TempPathFactory) -> None:
        conn = checks.get_db()
        cached = checks.get_cached_results(conn, [])
        assert cached == {}
        conn.close()

    def test_store_replaces_existing(self, tmp_path: pytest.TempPathFactory) -> None:
        conn = checks.get_db()
        checks.store_results(conn, 'msg@ex', [
            {'tool': 'lint', 'status': 'pass', 'summary': 'v1'}])
        checks.store_results(conn, 'msg@ex', [
            {'tool': 'lint', 'status': 'fail', 'summary': 'v2'}])
        cached = checks.get_cached_results(conn, ['msg@ex'])
        assert cached['msg@ex'][0]['status'] == 'fail'
        assert cached['msg@ex'][0]['summary'] == 'v2'
        conn.close()

    def test_delete_results(self, tmp_path: pytest.TempPathFactory) -> None:
        conn = checks.get_db()
        checks.store_results(conn, 'a@ex', [
            {'tool': 't1', 'status': 'pass'}])
        checks.store_results(conn, 'b@ex', [
            {'tool': 't1', 'status': 'pass'}])
        checks.delete_results(conn, ['a@ex'])
        cached = checks.get_cached_results(conn, ['a@ex', 'b@ex'])
        assert 'a@ex' not in cached
        assert 'b@ex' in cached
        conn.close()

    def test_delete_empty_list(self, tmp_path: pytest.TempPathFactory) -> None:
        conn = checks.get_db()
        checks.delete_results(conn, [])  # should not raise
        conn.close()

    def test_cleanup_old(self, tmp_path: pytest.TempPathFactory) -> None:
        conn = checks.get_db()
        checks.store_results(conn, 'recent@ex', [
            {'tool': 't', 'status': 'pass'}])
        # Manually backdate one row
        old_date = (datetime.datetime.now(datetime.timezone.utc)
                    - datetime.timedelta(days=200)).isoformat()
        conn.execute(
            "INSERT OR REPLACE INTO check_results"
            " (msgid, tool, status, checked_at)"
            " VALUES (?, ?, ?, ?)",
            ('old@ex', 't', 'pass', old_date))
        conn.commit()
        deleted = checks.cleanup_old(conn, max_days=180)
        assert deleted == 1
        cached = checks.get_cached_results(conn, ['recent@ex', 'old@ex'])
        assert 'recent@ex' in cached
        assert 'old@ex' not in cached
        conn.close()


# ---------------------------------------------------------------------------
# parse_cmd
# ---------------------------------------------------------------------------

class TestParseCmd:
    """Tests for parse_cmd shell splitting."""

    def test_simple(self) -> None:
        assert checks.parse_cmd('/usr/bin/check') == ['/usr/bin/check']

    def test_with_args(self) -> None:
        assert checks.parse_cmd('check --verbose -q') == [
            'check', '--verbose', '-q']

    def test_quoted_arg(self) -> None:
        assert checks.parse_cmd('check "hello world"') == [
            'check', 'hello world']

    def test_single_quotes(self) -> None:
        assert checks.parse_cmd("check 'hello world'") == [
            'check', 'hello world']


# ---------------------------------------------------------------------------
# _run_builtin_checkpatch output parsing
# ---------------------------------------------------------------------------

class TestBuiltinCheckpatch:
    """Tests for _run_builtin_checkpatch output parsing."""

    def _run(self, stdout: str, stderr: str = '',
             ecode: int = 0, topdir: str = '/fake') -> List[Dict[str, str]]:
        msg = _make_msg()
        with mock.patch('os.access', return_value=True), \
             mock.patch('b4._run_command', return_value=(
                 ecode,
                 stdout.encode() if stdout else b'',
                 stderr.encode() if stderr else b'')), \
             mock.patch('b4.LoreMessage.get_msg_as_bytes', return_value=b''):
            return checks._run_builtin_checkpatch(msg, topdir)

    def test_clean_pass(self) -> None:
        results = self._run('', ecode=0)
        assert len(results) == 1
        assert results[0]['status'] == 'pass'
        assert results[0]['tool'] == 'checkpatch'

    def test_error_lines(self) -> None:
        results = self._run('ERROR: trailing whitespace\n')
        assert results[0]['status'] == 'fail'
        assert '1 error' in results[0]['summary']

    def test_warning_lines(self) -> None:
        results = self._run('WARNING: missing Signed-off-by\n')
        assert results[0]['status'] == 'warn'
        assert '1 warning' in results[0]['summary']

    def test_check_treated_as_warning(self) -> None:
        results = self._run('CHECK: braces not needed\n')
        assert results[0]['status'] == 'warn'

    def test_mixed_errors_and_warnings(self) -> None:
        output = 'ERROR: bad thing\nWARNING: mild thing\nWARNING: another\n'
        results = self._run(output)
        assert results[0]['status'] == 'fail'
        assert '1 error' in results[0]['summary']
        assert '2 warnings' in results[0]['summary']

    def test_continuation_lines(self) -> None:
        output = 'WARNING: first part\n  continuation of warning\n'
        results = self._run(output)
        findings = json.loads(results[0]['details'])
        assert len(findings) == 1
        assert 'continuation' in findings[0]['description']

    def test_nonzero_exit_no_output(self) -> None:
        results = self._run('', ecode=1)
        assert results[0]['status'] == 'fail'
        assert 'error code' in results[0]['summary']

    def test_not_executable(self) -> None:
        msg = _make_msg()
        with mock.patch('os.access', return_value=False):
            results = checks._run_builtin_checkpatch(msg, '/fake')
        assert results[0]['status'] == 'fail'
        assert 'not found' in results[0]['summary']

    def test_dash_prefix_stripped(self) -> None:
        results = self._run('-:42: WARNING: something bad\n')
        findings = json.loads(results[0]['details'])
        # The leading "-:" should be stripped
        assert not findings[0]['description'].startswith('-:')


# ---------------------------------------------------------------------------
# _run_external_cmd JSON protocol
# ---------------------------------------------------------------------------

class TestRunExternalCmd:
    """Tests for _run_external_cmd JSON parsing."""

    def _run(self, stdout: str, stderr: str = '',
             ecode: int = 0) -> List[Dict[str, str]]:
        msg = _make_msg()
        with mock.patch('b4._run_command', return_value=(
                ecode,
                stdout.encode() if stdout else b'',
                stderr.encode() if stderr else b'')), \
             mock.patch('b4.LoreMessage.get_msg_as_bytes', return_value=b''):
            return checks._run_external_cmd(['mycheck'], msg, '/fake')

    def test_valid_json_array(self) -> None:
        data = [{'tool': 'ci', 'status': 'pass', 'summary': 'ok'}]
        results = self._run(json.dumps(data))
        assert len(results) == 1
        assert results[0]['tool'] == 'ci'
        assert results[0]['status'] == 'pass'

    def test_single_object_wrapped(self) -> None:
        data = {'tool': 'ci', 'status': 'warn', 'summary': 'hmm'}
        results = self._run(json.dumps(data))
        assert len(results) == 1
        assert results[0]['status'] == 'warn'

    def test_invalid_json(self) -> None:
        results = self._run('not json at all')
        assert len(results) == 1
        assert results[0]['status'] == 'fail'
        assert 'invalid JSON' in results[0]['summary']

    def test_empty_output_zero_exit(self) -> None:
        results = self._run('')
        assert results == []

    def test_empty_output_nonzero_exit(self) -> None:
        results = self._run('', ecode=1)
        assert len(results) == 1
        assert results[0]['status'] == 'fail'
        assert 'error code' in results[0]['summary']

    def test_invalid_status_defaults_to_fail(self) -> None:
        data = [{'tool': 'ci', 'status': 'banana'}]
        results = self._run(json.dumps(data))
        assert results[0]['status'] == 'fail'

    def test_missing_tool_uses_basename(self) -> None:
        data = [{'status': 'pass'}]
        results = self._run(json.dumps(data))
        assert results[0]['tool'] == 'mycheck'

    def test_non_dict_entries_skipped(self) -> None:
        data = [{'tool': 'ci', 'status': 'pass'}, 'garbage', 42]
        results = self._run(json.dumps(data))
        assert len(results) == 1

    def test_optional_fields_default_empty(self) -> None:
        data = [{'tool': 'ci', 'status': 'pass'}]
        results = self._run(json.dumps(data))
        assert results[0]['summary'] == ''
        assert results[0]['url'] == ''
        assert results[0]['details'] == ''

    def test_stderr_in_error_details(self) -> None:
        results = self._run('', stderr='something broke', ecode=1)
        assert 'something broke' in results[0]['details']

    def test_extra_env_set_during_run(self) -> None:
        captured_env: Dict[str, str] = {}

        def fake_run(cmdargs: Any, stdin: Any = None,
                     rundir: Any = None) -> Any:
            captured_env['B4_TRACKING_FILE'] = os.environ.get('B4_TRACKING_FILE', '')
            return (0, b'[]', b'')

        msg = _make_msg()
        with mock.patch('b4._run_command', side_effect=fake_run), \
             mock.patch('b4.LoreMessage.get_msg_as_bytes', return_value=b''):
            checks._run_external_cmd(['mycheck'], msg, '/fake',
                                     extra_env={'B4_TRACKING_FILE': '/tmp/test.json'})
        assert captured_env['B4_TRACKING_FILE'] == '/tmp/test.json'
        # Env var should be cleaned up after the call
        assert 'B4_TRACKING_FILE' not in os.environ

    def test_extra_env_restored_on_error(self) -> None:
        msg = _make_msg()
        os.environ['B4_TRACKING_FILE'] = 'original'
        try:
            with mock.patch('b4._run_command', side_effect=RuntimeError('boom')), \
                 mock.patch('b4.LoreMessage.get_msg_as_bytes', return_value=b''):
                try:
                    checks._run_external_cmd(
                        ['mycheck'], msg, '/fake',
                        extra_env={'B4_TRACKING_FILE': '/tmp/new.json'})
                except RuntimeError:
                    pass
            assert os.environ.get('B4_TRACKING_FILE') == 'original'
        finally:
            os.environ.pop('B4_TRACKING_FILE', None)


# ---------------------------------------------------------------------------
# _run_builtin_patchwork aggregation
# ---------------------------------------------------------------------------

class TestBuiltinPatchwork:
    """Tests for _run_builtin_patchwork status aggregation."""

    def _run(self, pw_checks: List[Dict[str, Any]]) -> List[Dict[str, str]]:
        msg = _make_msg(msgid='test@example.com')
        with mock.patch('b4.LoreMessage.get_patchwork_data_by_msgid',
                        return_value={'id': 42}), \
             mock.patch('b4.review.pw_fetch_checks',
                        return_value=pw_checks):
            return checks._run_builtin_patchwork(msg, 'proj', 'https://pw.example.com')

    def test_all_success(self) -> None:
        pw = [
            {'state': 'success', 'context': 'build', 'description': 'ok', 'url': ''},
            {'state': 'success', 'context': 'test', 'description': 'ok', 'url': ''},
        ]
        results = self._run(pw)
        assert len(results) == 1
        assert results[0]['tool'] == 'patchwork'
        assert results[0]['status'] == 'pass'

    def test_worst_case_fail(self) -> None:
        pw = [
            {'state': 'success', 'context': 'build', 'description': 'ok', 'url': ''},
            {'state': 'fail', 'context': 'test', 'description': 'bad', 'url': ''},
        ]
        results = self._run(pw)
        assert results[0]['status'] == 'fail'

    def test_pending_is_warn(self) -> None:
        pw = [
            {'state': 'pending', 'context': 'ci', 'description': 'running', 'url': ''},
        ]
        results = self._run(pw)
        assert results[0]['status'] == 'warn'

    def test_warning_is_warn(self) -> None:
        pw = [
            {'state': 'warning', 'context': 'ci', 'description': 'iffy', 'url': ''},
        ]
        results = self._run(pw)
        assert results[0]['status'] == 'warn'

    def test_details_are_json(self) -> None:
        pw = [
            {'state': 'success', 'context': 'build', 'description': 'ok', 'url': 'http://x'},
        ]
        results = self._run(pw)
        details = json.loads(results[0]['details'])
        assert isinstance(details, list)
        assert details[0]['context'] == 'build'

    def test_no_msgid_returns_empty(self) -> None:
        msg = EmailMessage()
        msg['Subject'] = 'test'
        result = checks._run_builtin_patchwork(msg, 'proj', 'https://pw.example.com')
        assert result == []

    def test_lookup_failure_returns_empty(self) -> None:
        msg = _make_msg()
        with mock.patch('b4.LoreMessage.get_patchwork_data_by_msgid',
                        side_effect=LookupError('not found')):
            result = checks._run_builtin_patchwork(msg, 'proj', 'https://pw.example.com')
        assert result == []


# ---------------------------------------------------------------------------
# High-level runners
# ---------------------------------------------------------------------------

class TestRunners:
    """Tests for run_perpatch_checks and run_series_checks."""

    def test_perpatch_dispatches_external(self) -> None:
        msg = _make_msg()
        data = json.dumps([{'tool': 'ci', 'status': 'pass', 'summary': 'ok'}])
        with mock.patch('b4._run_command', return_value=(
                0, data.encode(), b'')), \
             mock.patch('b4.LoreMessage.get_msg_as_bytes', return_value=b''):
            results = checks.run_perpatch_checks(
                [('m1@ex', msg)], ['mycheck'], '/fake')
        assert 'm1@ex' in results
        assert results['m1@ex'][0]['tool'] == 'ci'

    def test_perpatch_exception_captured(self) -> None:
        msg = _make_msg()
        with mock.patch('b4._run_command',
                        side_effect=RuntimeError('boom')), \
             mock.patch('b4.LoreMessage.get_msg_as_bytes', return_value=b''):
            results = checks.run_perpatch_checks(
                [('m1@ex', msg)], ['badcmd'], '/fake')
        assert results['m1@ex'][0]['status'] == 'fail'
        assert 'boom' in results['m1@ex'][0]['summary']

    def test_series_dispatches_external(self) -> None:
        msg = _make_msg()
        data = json.dumps([{'tool': 'series-ci', 'status': 'warn'}])
        with mock.patch('b4._run_command', return_value=(
                0, data.encode(), b'')), \
             mock.patch('b4.LoreMessage.get_msg_as_bytes', return_value=b''):
            results = checks.run_series_checks(
                ('cover@ex', msg), ['mycheck'], '/fake')
        assert len(results) == 1
        assert results[0]['tool'] == 'series-ci'

    def test_series_exception_captured(self) -> None:
        msg = _make_msg()
        with mock.patch('b4._run_command',
                        side_effect=RuntimeError('kaboom')), \
             mock.patch('b4.LoreMessage.get_msg_as_bytes', return_value=b''):
            results = checks.run_series_checks(
                ('cover@ex', msg), ['badcmd'], '/fake')
        assert results[0]['status'] == 'fail'
        assert 'kaboom' in results[0]['summary']

    def test_dispatch_builtin_checkpatch(self) -> None:
        msg = _make_msg()
        with mock.patch('os.access', return_value=True), \
             mock.patch('b4._run_command', return_value=(0, b'', b'')), \
             mock.patch('b4.LoreMessage.get_msg_as_bytes', return_value=b''):
            results = checks._dispatch_cmd('_builtin_checkpatch', msg, '/fake')
        assert results[0]['tool'] == 'checkpatch'

    def test_dispatch_builtin_patchwork_without_config(self) -> None:
        msg = _make_msg()
        results = checks._dispatch_cmd('_builtin_patchwork', msg, '/fake')
        assert results == []


# ---------------------------------------------------------------------------
# _STATUS_ORDER module-level constant
# ---------------------------------------------------------------------------

class TestStatusOrder:
    """Verify the module-level status ordering constant."""

    def test_ordering(self) -> None:
        assert checks._STATUS_ORDER['pass'] < checks._STATUS_ORDER['warn']
        assert checks._STATUS_ORDER['warn'] < checks._STATUS_ORDER['fail']


# ---------------------------------------------------------------------------
# Sashiko AI review integration
# ---------------------------------------------------------------------------

# Sample patchset response matching the real sashiko API format.
_SASHIKO_PATCHSET: Dict[str, Any] = {
    'id': 93,
    'message_id': 'cover@example.com',
    'subject': '[PATCH 0/3] Example series',
    'status': 'Reviewed',
    'author': 'Test Author <test@example.com>',
    'patches': [
        {'id': 1, 'message_id': 'patch1@example.com', 'part_index': 1,
         'subject': '[PATCH 1/3] First patch', 'status': 'applied'},
        {'id': 2, 'message_id': 'patch2@example.com', 'part_index': 2,
         'subject': '[PATCH 2/3] Second patch', 'status': 'applied'},
        {'id': 3, 'message_id': 'patch3@example.com', 'part_index': 3,
         'subject': '[PATCH 3/3] Third patch', 'status': 'applied'},
    ],
    'reviews': [
        {
            'id': 100, 'patch_id': 1, 'status': 'Reviewed',
            'result': 'Review completed successfully.',
            'summary': '', 'inline_review': 'looks good',
            'output': json.dumps({
                'findings': [
                    {'severity': 'Low', 'problem': 'Minor style issue'},
                ],
            }),
        },
        {
            'id': 101, 'patch_id': 2, 'status': 'Reviewed',
            'result': 'Review completed successfully.',
            'summary': '', 'inline_review': 'has issues',
            'output': json.dumps({
                'findings': [
                    {'severity': 'Critical', 'problem': 'Use-after-free',
                     'suggestion': 'Add proper locking'},
                    {'severity': 'High', 'problem': 'Missing error check'},
                ],
            }),
        },
        {
            'id': 102, 'patch_id': 3, 'status': 'Skipped',
            'result': 'Skipped: touches only ignored files',
            'summary': '', 'inline_review': '', 'output': '',
        },
    ],
}


class TestSashikoCache:
    """Tests for sashiko in-process patchset cache."""

    def setup_method(self) -> None:
        checks.clear_sashiko_cache()

    def teardown_method(self) -> None:
        checks.clear_sashiko_cache()

    def test_clear_cache(self) -> None:
        checks._sashiko_patchset_cache['test@ex'] = {'id': 1}
        checks.clear_sashiko_cache()
        assert checks._sashiko_patchset_cache == {}

    def test_fetch_caches_all_msgids(self) -> None:
        resp = mock.Mock()
        resp.status_code = 200
        resp.json.return_value = _SASHIKO_PATCHSET

        session = mock.Mock()
        session.get.return_value = resp

        with mock.patch('b4.get_requests_session', return_value=session):
            data = checks._fetch_sashiko_patchset(
                'cover@example.com', 'https://sashiko.dev')

        assert data is not None
        assert data['id'] == 93
        # All msgids should be cached
        assert 'cover@example.com' in checks._sashiko_patchset_cache
        assert 'patch1@example.com' in checks._sashiko_patchset_cache
        assert 'patch2@example.com' in checks._sashiko_patchset_cache
        assert 'patch3@example.com' in checks._sashiko_patchset_cache
        # Second call should use cache, not network
        session.get.reset_mock()
        data2 = checks._fetch_sashiko_patchset(
            'patch2@example.com', 'https://sashiko.dev')
        session.get.assert_not_called()
        assert data2 is not None
        assert data2['id'] == 93

    def test_fetch_404_caches_none(self) -> None:
        resp = mock.Mock()
        resp.status_code = 404

        session = mock.Mock()
        session.get.return_value = resp

        with mock.patch('b4.get_requests_session', return_value=session):
            data = checks._fetch_sashiko_patchset(
                'unknown@example.com', 'https://sashiko.dev')

        assert data is None
        assert checks._sashiko_patchset_cache['unknown@example.com'] is None

    def test_fetch_network_error_caches_none(self) -> None:
        import requests
        session = mock.Mock()
        session.get.side_effect = requests.ConnectionError('offline')

        with mock.patch('b4.get_requests_session', return_value=session):
            data = checks._fetch_sashiko_patchset(
                'test@example.com', 'https://sashiko.dev')

        assert data is None
        assert checks._sashiko_patchset_cache['test@example.com'] is None


class TestParseSashikoFindings:
    """Tests for _parse_sashiko_findings."""

    def test_empty_output(self) -> None:
        assert checks._parse_sashiko_findings({'output': ''}) == []

    def test_null_output(self) -> None:
        assert checks._parse_sashiko_findings({'output': None}) == []

    def test_no_output_key(self) -> None:
        assert checks._parse_sashiko_findings({}) == []

    def test_invalid_json(self) -> None:
        assert checks._parse_sashiko_findings({'output': 'not json'}) == []

    def test_critical_finding(self) -> None:
        review = {'output': json.dumps({
            'findings': [{'severity': 'Critical', 'problem': 'UAF bug'}],
        })}
        findings = checks._parse_sashiko_findings(review)
        assert len(findings) == 1
        assert findings[0]['status'] == 'fail'
        assert findings[0]['state'] == 'critical'
        assert 'UAF bug' in findings[0]['description']

    def test_high_finding(self) -> None:
        review = {'output': json.dumps({
            'findings': [{'severity': 'High', 'problem': 'Missing check'}],
        })}
        findings = checks._parse_sashiko_findings(review)
        assert findings[0]['status'] == 'fail'
        assert findings[0]['state'] == 'high'

    def test_medium_finding(self) -> None:
        review = {'output': json.dumps({
            'findings': [{'severity': 'Medium', 'problem': 'Questionable logic'}],
        })}
        findings = checks._parse_sashiko_findings(review)
        assert findings[0]['status'] == 'warn'
        assert findings[0]['state'] == 'medium'

    def test_low_finding(self) -> None:
        review = {'output': json.dumps({
            'findings': [{'severity': 'Low', 'problem': 'Style issue'}],
        })}
        findings = checks._parse_sashiko_findings(review)
        assert findings[0]['status'] == 'pass'
        assert findings[0]['state'] == 'low'

    def test_suggestion_appended(self) -> None:
        review = {'output': json.dumps({
            'findings': [{'severity': 'High', 'problem': 'Bug',
                         'suggestion': 'Fix it'}],
        })}
        findings = checks._parse_sashiko_findings(review)
        assert 'Bug' in findings[0]['description']
        assert 'Fix it' in findings[0]['description']

    def test_context_includes_severity(self) -> None:
        review = {'output': json.dumps({
            'findings': [{'severity': 'Medium', 'problem': 'test'}],
        })}
        findings = checks._parse_sashiko_findings(review)
        assert findings[0]['context'] == 'sashiko/medium'

    def test_multiple_findings(self) -> None:
        review = {'output': json.dumps({
            'findings': [
                {'severity': 'Critical', 'problem': 'bad'},
                {'severity': 'Low', 'problem': 'minor'},
            ],
        })}
        findings = checks._parse_sashiko_findings(review)
        assert len(findings) == 2

    def test_no_findings_key(self) -> None:
        review = {'output': json.dumps({'fixes': []})}
        assert checks._parse_sashiko_findings(review) == []


class TestSashikoFindingsSummary:
    """Tests for _sashiko_findings_summary."""

    def test_no_findings(self) -> None:
        worst, summary = checks._sashiko_findings_summary([])
        assert worst == 'pass'
        assert summary == 'No findings'

    def test_single_critical(self) -> None:
        findings = [{'status': 'fail', 'state': 'critical',
                     'description': 'bad'}]
        worst, summary = checks._sashiko_findings_summary(findings)
        assert worst == 'fail'
        assert '1 critical' in summary

    def test_mixed_severities(self) -> None:
        findings = [
            {'status': 'fail', 'state': 'critical', 'description': ''},
            {'status': 'fail', 'state': 'high', 'description': ''},
            {'status': 'warn', 'state': 'medium', 'description': ''},
            {'status': 'pass', 'state': 'low', 'description': ''},
        ]
        worst, summary = checks._sashiko_findings_summary(findings)
        assert worst == 'fail'
        assert '1 critical' in summary
        assert '1 high' in summary
        assert '1 medium' in summary
        assert '1 low' in summary

    def test_only_low_is_pass(self) -> None:
        findings = [
            {'status': 'pass', 'state': 'low', 'description': ''},
            {'status': 'pass', 'state': 'low', 'description': ''},
        ]
        worst, summary = checks._sashiko_findings_summary(findings)
        assert worst == 'pass'
        assert '2 low' in summary


class TestRunBuiltinSashiko:
    """Tests for _run_builtin_sashiko end-to-end."""

    def setup_method(self) -> None:
        checks.clear_sashiko_cache()

    def teardown_method(self) -> None:
        checks.clear_sashiko_cache()

    def _prefill_cache(self, patchset: Optional[Dict[str, Any]] = None) -> None:
        """Pre-fill the cache so no HTTP calls are made."""
        ps = patchset if patchset is not None else _SASHIKO_PATCHSET
        for key in ['cover@example.com', 'patch1@example.com',
                    'patch2@example.com', 'patch3@example.com']:
            checks._sashiko_patchset_cache[key] = ps

    def test_no_msgid_returns_empty(self) -> None:
        msg = EmailMessage()
        msg['Subject'] = 'test'
        assert checks._run_builtin_sashiko(msg, 'https://sashiko.dev') == []

    def test_not_found_returns_empty(self) -> None:
        checks._sashiko_patchset_cache['unknown@ex'] = None
        msg = _make_msg(msgid='unknown@ex')
        assert checks._run_builtin_sashiko(msg, 'https://sashiko.dev') == []

    def test_cover_letter_aggregates_all_findings(self) -> None:
        self._prefill_cache()
        msg = _make_msg(msgid='cover@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert len(results) == 1
        assert results[0]['tool'] == 'sashiko'
        assert results[0]['status'] == 'fail'  # critical finding in patch 2
        assert '1 critical' in results[0]['summary']
        assert '1 high' in results[0]['summary']
        assert '1 low' in results[0]['summary']
        assert results[0]['url'] == 'https://sashiko.dev/patch/93'
        # Details should be valid JSON
        details = json.loads(results[0]['details'])
        assert len(details) == 3  # 1 low + 1 critical + 1 high

    def test_patch_with_critical_finding(self) -> None:
        self._prefill_cache()
        msg = _make_msg(msgid='patch2@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'fail'
        assert '1 critical' in results[0]['summary']
        assert '1 high' in results[0]['summary']

    def test_patch_with_low_finding(self) -> None:
        self._prefill_cache()
        msg = _make_msg(msgid='patch1@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'pass'
        assert '1 low' in results[0]['summary']

    def test_skipped_patch(self) -> None:
        self._prefill_cache()
        msg = _make_msg(msgid='patch3@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'pass'
        assert 'Skipped' in results[0]['summary']

    def test_pending_patchset(self) -> None:
        ps = dict(_SASHIKO_PATCHSET, status='Pending')
        self._prefill_cache(ps)
        msg = _make_msg(msgid='patch1@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'warn'
        assert 'pending' in results[0]['summary'].lower()

    def test_in_review_patchset(self) -> None:
        ps = dict(_SASHIKO_PATCHSET, status='In Review')
        self._prefill_cache(ps)
        msg = _make_msg(msgid='cover@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'warn'
        assert 'in review' in results[0]['summary'].lower()

    def test_failed_patchset(self) -> None:
        ps = dict(_SASHIKO_PATCHSET, status='Failed')
        self._prefill_cache(ps)
        msg = _make_msg(msgid='cover@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'fail'
        assert results[0]['summary'] == 'Failed'

    def test_failed_to_apply_patchset(self) -> None:
        ps = dict(_SASHIKO_PATCHSET, status='Failed To Apply')
        self._prefill_cache(ps)
        msg = _make_msg(msgid='cover@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'fail'

    def test_incomplete_patchset(self) -> None:
        ps = dict(_SASHIKO_PATCHSET, status='Incomplete')
        self._prefill_cache(ps)
        msg = _make_msg(msgid='cover@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'warn'
        assert 'incomplete' in results[0]['summary'].lower()

    def test_no_findings_pass(self) -> None:
        reviews = [{'id': 100, 'patch_id': 1, 'status': 'Reviewed',
                    'result': 'Review completed successfully.',
                    'output': json.dumps({'findings': []})}]
        ps = dict(_SASHIKO_PATCHSET, reviews=reviews)
        self._prefill_cache(ps)
        msg = _make_msg(msgid='patch1@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'pass'
        assert results[0]['summary'] == 'No findings'

    def test_pending_review_for_patch(self) -> None:
        reviews = [{'id': 100, 'patch_id': 1, 'status': 'Pending',
                    'output': ''}]
        ps = dict(_SASHIKO_PATCHSET, reviews=reviews)
        self._prefill_cache(ps)
        msg = _make_msg(msgid='patch1@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'warn'
        assert 'in progress' in results[0]['summary'].lower()

    def test_failed_review_for_patch(self) -> None:
        reviews = [{'id': 100, 'patch_id': 1, 'status': 'Failed',
                    'result': 'Token limit exceeded', 'output': ''}]
        ps = dict(_SASHIKO_PATCHSET, reviews=reviews)
        self._prefill_cache(ps)
        msg = _make_msg(msgid='patch1@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'fail'
        assert 'Token limit' in results[0]['summary']

    def test_patch_not_in_sashiko(self) -> None:
        self._prefill_cache()
        msg = _make_msg(msgid='unknown-patch@example.com')
        # Not in cache, will try to fetch
        checks._sashiko_patchset_cache['unknown-patch@example.com'] = None
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results == []

    def test_patch_without_review(self) -> None:
        # Patchset is reviewed but this specific patch has no review entry
        ps = dict(_SASHIKO_PATCHSET, reviews=[])
        self._prefill_cache(ps)
        msg = _make_msg(msgid='patch1@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev')
        assert results[0]['status'] == 'pass'
        assert results[0]['summary'] == 'No review'

    def test_url_constructed_correctly(self) -> None:
        self._prefill_cache()
        msg = _make_msg(msgid='patch1@example.com')
        results = checks._run_builtin_sashiko(msg, 'https://sashiko.dev/')
        # Trailing slash should not cause double slash
        assert results[0]['url'] == 'https://sashiko.dev/patch/93'


class TestSashikoAutoWire:
    """Tests for auto-wiring _builtin_sashiko in load_check_cmds."""

    def test_sashiko_added_when_url_configured(self) -> None:
        config = {'sashiko-url': 'https://sashiko.dev'}
        with mock.patch('b4.get_main_config', return_value=config), \
             mock.patch('b4.git_get_toplevel', return_value=None):
            perpatch, series = checks.load_check_cmds()
        assert '_builtin_sashiko' in perpatch
        assert '_builtin_sashiko' in series

    def test_sashiko_not_added_without_url(self) -> None:
        config: Dict[str, Any] = {}
        with mock.patch('b4.get_main_config', return_value=config), \
             mock.patch('b4.git_get_toplevel', return_value=None):
            perpatch, series = checks.load_check_cmds()
        assert '_builtin_sashiko' not in perpatch
        assert '_builtin_sashiko' not in series

    def test_sashiko_not_duplicated(self) -> None:
        config = {
            'sashiko-url': 'https://sashiko.dev',
            'review-perpatch-check-cmd': ['_builtin_sashiko'],
            'review-series-check-cmd': ['_builtin_sashiko'],
        }
        with mock.patch('b4.get_main_config', return_value=config), \
             mock.patch('b4.git_get_toplevel', return_value=None):
            perpatch, series = checks.load_check_cmds()
        assert perpatch.count('_builtin_sashiko') == 1
        assert series.count('_builtin_sashiko') == 1


class TestSashikoDispatch:
    """Tests for _dispatch_cmd routing to _builtin_sashiko."""

    def test_dispatch_routes_to_sashiko(self) -> None:
        checks.clear_sashiko_cache()
        checks._sashiko_patchset_cache['test@ex'] = _SASHIKO_PATCHSET
        msg = _make_msg(msgid='test@ex')
        # Pre-cache so no HTTP call is made; use cover msgid
        checks._sashiko_patchset_cache['test@ex'] = dict(
            _SASHIKO_PATCHSET, message_id='test@ex')
        config = {'sashiko-url': 'https://sashiko.dev'}
        with mock.patch('b4.get_main_config', return_value=config):
            results = checks._dispatch_cmd('_builtin_sashiko', msg, '/fake')
        assert results[0]['tool'] == 'sashiko'
        checks.clear_sashiko_cache()

    def test_dispatch_without_config(self) -> None:
        msg = _make_msg()
        config: Dict[str, Any] = {}
        with mock.patch('b4.get_main_config', return_value=config):
            results = checks._dispatch_cmd('_builtin_sashiko', msg, '/fake')
        assert results == []



import pytest
import b4
import os
import email
import email.parser
import io
import pathlib
import socket

from typing import Any, Dict, List, Literal, Optional, Tuple


@pytest.mark.parametrize('source,expected', [
    ('good-valid-trusted', (True, True, True, 'B6C41CE35664996C', '1623274836')),
    ('good-valid-notrust', (True, True, False, 'B6C41CE35664996C', '1623274836')),
    ('good-invalid-notrust', (True, False, False, 'B6C41CE35664996C', None)),
    ('badsig', (False, False, False, 'B6C41CE35664996C', None)),
    ('no-pubkey', (False, False, False, None, None)),
])
def test_check_gpg_status(sampledir: str, source: str, expected: Tuple[bool, bool, bool, Optional[str], Optional[str]]) -> None:
    with open(f'{sampledir}/gpg-{source}.txt', 'r') as fh:
        status = fh.read()
    assert b4.check_gpg_status(status) == expected


@pytest.mark.parametrize('source,regex,flags,ismbox', [
    (None, r'^From git@z ', 0, False),
    (None, r'\n\nFrom git@z ', 0, False),
    ('save-7bit-clean', r'From: Unicôdé', 0, True),
    # mailbox.mbox does not properly handle 8bit-clean headers
    ('save-8bit-clean', r'From: Unicôdé', 0, False),
])
def test_save_git_am_mbox(sampledir: Optional[str], tmp_path: pathlib.Path, source: Optional[str], regex: str, flags: int, ismbox: bool) -> None:
    import re
    msgs: List[email.message.EmailMessage]
    if source is not None:
        if ismbox:
            msgs = b4.get_msgs_from_mailbox_or_maildir(f'{sampledir}/{source}.txt')
        else:
            import email
            import email.parser
            with open(f'{sampledir}/{source}.txt', 'rb') as fh:
                msg = email.parser.BytesParser(policy=b4.emlpolicy, _class=email.message.EmailMessage).parse(fh)
            msgs = [msg]
    else:
        import email.message
        msgs = list()
        for x in range(0, 3):
            msg = email.message.EmailMessage()
            msg.set_payload(f'Hello world {x}\n')
            msg['Subject'] = f'Hello world {x}'
            msg['From'] = f'Me{x} <me{x}@foo.bar>'
            msgs.append(msg)
    dest = os.path.join(tmp_path, 'out')
    with open(dest, 'wb') as fh:
        b4.save_git_am_mbox(msgs, fh)
    with open(dest, 'r') as fh:
        res = fh.read()
    assert re.search(regex, res, flags=flags)


def _msgid_domain(msgid: str) -> str:
    return msgid.strip('<>').rsplit('@', maxsplit=1)[1]


def test_make_msgid_avoids_host_domain_by_default() -> None:
    stdlib_msgid = email.utils.make_msgid()
    b4_msgid = b4.make_msgid(idstring='b4-test')

    assert _msgid_domain(stdlib_msgid) == socket.getfqdn()
    assert _msgid_domain(b4_msgid) == 'b4'
    assert _msgid_domain(b4_msgid) != socket.getfqdn()


@pytest.mark.parametrize('source,expected', [
    ('trailers-test-simple',
     [('person', 'Reported-by', '"Doe, Jane" <jane@example.com>', None),
      ('person', 'Reviewed-by', 'Bogus Bupkes <bogus@example.com>', None),
      ('utility', 'Fixes', 'abcdef01234567890', None),
      ('utility', 'Link', 'https://msgid.link/some@msgid.here', None),
      ]),
    ('trailers-test-extinfo',
     [('person', 'Reported-by', 'Some, One <somewhere@example.com>', None),
      ('person', 'Reviewed-by', 'Bogus Bupkes <bogus@example.com>', '[for the parts that are bogus]'),
      ('utility', 'Fixes', 'abcdef01234567890', None),
      ('person', 'Tested-by', 'Some Person <bogus2@example.com>', '           [this person visually indented theirs]'),
      ('utility', 'Link', 'https://msgid.link/some@msgid.here', '  # initial submission'),
      ('person', 'Signed-off-by', 'Wrapped Persontrailer <broken@example.com>', None),
      ]),
])
def test_parse_trailers(sampledir: str, source: str, expected: List[Tuple[str, str, str, Optional[str]]]) -> None:
    msgs = b4.get_msgs_from_mailbox_or_maildir(f'{sampledir}/{source}.txt')
    for msg in msgs:
        lmsg = b4.LoreMessage(msg)
        _, _, trs, _, _ = b4.LoreMessage.get_body_parts(lmsg.body)
        assert len(expected) == len(trs)
        for tr in trs:
            mytype, myname, myvalue, myextinfo = expected.pop(0)
            assert tr.name == myname
            assert tr.value == myvalue
            assert tr.extinfo == myextinfo
            assert tr.type == mytype

            mytr = b4.LoreTrailer(name=myname, value=myvalue, extinfo=myextinfo)
            assert tr == mytr
            assert tr.extinfo == mytr.extinfo


@pytest.mark.parametrize('name,value,exp_type,exp_addr,exp_value', [
    # Simple name
    ('Signed-off-by', 'Simple Name <simple@example.com>',
     'person', ('Simple Name', 'simple@example.com'),
     'Simple Name <simple@example.com>'),
    # Double quotes in display name must be preserved
    ('Signed-off-by', 'Jane "JD" Doe <jd@example.com>',
     'person', ('Jane "JD" Doe', 'jd@example.com'),
     'Jane "JD" Doe <jd@example.com>'),
    # Outer RFC 2822 quotes around a name with comma
    ('Reported-by', '"Doe, Jane" <jane@example.com>',
     'person', ('"Doe, Jane"', 'jane@example.com'),
     '"Doe, Jane" <jane@example.com>'),
    # Comma in name without quotes
    ('Reported-by', 'Some, One <somewhere@example.com>',
     'person', ('Some, One', 'somewhere@example.com'),
     'Some, One <somewhere@example.com>'),
    # Parentheses in display name
    ('Tested-by', 'Developer Foo (EXAMPLECORP) <dev@example.com>',
     'person', ('Developer Foo (EXAMPLECORP)', 'dev@example.com'),
     'Developer Foo (EXAMPLECORP) <dev@example.com>'),
    # Bare angle-bracket email
    ('Cc', '<bare@example.com>',
     'person', ('', 'bare@example.com'),
     'bare@example.com'),
    # Bare email without angle brackets
    ('Cc', 'bare@example.com',
     'person', ('', 'bare@example.com'),
     'bare@example.com'),
])
def test_trailer_addr_parsing(name: str, value: str, exp_type: str,
                              exp_addr: Tuple[str, str], exp_value: str) -> None:
    tr = b4.LoreTrailer(name=name, value=value)
    assert tr.type == exp_type
    assert tr.addr == exp_addr
    assert tr.value == exp_value


@pytest.mark.parametrize('source,serargs,amargs,reference,b4cfg', [
    ('single', {}, {}, 'defaults', {}),
    ('single', {}, {'noaddtrailers': True}, 'noadd', {}),
    ('single', {}, {'addmysob': True}, 'addmysob', {}),
    ('single', {}, {'addmysob': True, 'copyccs': True}, 'copyccs', {}),
    ('single', {}, {'addmysob': True, 'addlink': True}, 'addlink', {}),
    ('single', {}, {'addmysob': True, 'addlink': True}, 'addmsgid', {'linktrailermask': 'Message-ID: <%s>'}),
    ('single', {}, {'addmysob': True, 'copyccs': True}, 'ordered',
     {'trailer-order': 'Cc,Tested*,Reviewed*,*'}),
    ('single', {'sloppytrailers': True}, {'addmysob': True}, 'sloppy', {}),
    ('with-cover', {}, {'addmysob': True}, 'defaults', {}),
    ('with-cover', {}, {'addmysob': True, 'addlink': True}, 'addlink', {}),
    ('custody', {}, {'addmysob': True, 'copyccs': True}, 'unordered', {}),
    ('custody', {}, {'addmysob': True, 'copyccs': True}, 'ordered',
     {'trailer-order': 'Cc,Fixes*,Link*,Suggested*,Reviewed*,Tested*,*'}),
    ('custody', {}, {'addmysob': True, 'copyccs': True}, 'with-ignored',
     {'trailers-ignore-from': 'followup-reviewer1@example.com'}),
    ('partial-reroll', {}, {'addmysob': True}, 'defaults', {}),
    ('nore', {}, {}, 'defaults', {}),
    ('non-git-patch', {}, {}, 'defaults', {}),
    ('non-git-patch-with-comments', {}, {}, 'defaults', {}),
    ('with-diffstat', {}, {}, 'defaults', {}),
    ('name-parens', {}, {}, 'defaults', {}),
    ('bare-address', {}, {}, 'defaults', {}),
    ('stripped-lines', {}, {}, 'defaults', {}),
    ('htmljunk', {}, {}, 'defaults', {}),
])
def test_followup_trailers(sampledir: str, source: str, serargs: Dict[str, Any], amargs: Dict[str, Any],
                           reference: str, b4cfg: Dict[str, Any]) -> None:
    b4.MAIN_CONFIG.update(b4cfg)
    lmbx = b4.LoreMailbox()
    for msg in b4.get_msgs_from_mailbox_or_maildir(f'{sampledir}/trailers-followup-{source}.mbox'):
        lmbx.add_message(msg)
    lser = lmbx.get_series(**serargs)
    assert lser is not None
    amsgs = lser.get_am_ready(**amargs)
    ifh = io.BytesIO()
    b4.save_git_am_mbox(amsgs, ifh)
    with open(f'{sampledir}/trailers-followup-{source}-ref-{reference}.txt', 'r') as fh:
        assert ifh.getvalue().decode() == fh.read()


@pytest.mark.parametrize('hval,verify,tr', [
    ('short-ascii', 'short-ascii', 'encode'),
    ('short-unicôde', '=?utf-8?q?short-unic=C3=B4de?=', 'encode'),
    # Long ascii
    (('Lorem ipsum dolor sit amet consectetur adipiscing elit '
      'sed do eiusmod tempor incididunt ut labore et dolore magna aliqua'),
     ('Lorem ipsum dolor sit amet consectetur adipiscing elit sed do\n'
      ' eiusmod tempor incididunt ut labore et dolore magna aliqua'), 'encode'),
    # Long unicode
    (('Lorem îpsum dolor sit amet consectetur adipiscing elît '
      'sed do eiusmod tempôr incididunt ut labore et dolôre magna aliqua'),
     ('=?utf-8?q?Lorem_=C3=AEpsum_dolor_sit_amet_consectetur_adipiscin?=\n'
      ' =?utf-8?q?g_el=C3=AEt_sed_do_eiusmod_temp=C3=B4r_incididunt_ut_labore_et?=\n'
      ' =?utf-8?q?_dol=C3=B4re_magna_aliqua?='), 'encode'),
    # Exactly 75 long
    ('Lorem ipsum dolor sit amet consectetur adipiscing elit sed do eiu',
     'Lorem ipsum dolor sit amet consectetur adipiscing elit sed do eiu', 'encode'),
    # Unicode that breaks on escape boundary
    ('Lorem ipsum dolor sit amet consectetur adipiscin elît',
     '=?utf-8?q?Lorem_ipsum_dolor_sit_amet_consectetur_adipiscin_el?=\n =?utf-8?q?=C3=AEt?=', 'encode'),
    # Unicode that's just 1 too long
    ('Lorem ipsum dolor sit amet consectetur adipi elît',
     '=?utf-8?q?Lorem_ipsum_dolor_sit_amet_consectetur_adipi_el=C3=AE?=\n =?utf-8?q?t?=', 'encode'),
    # A single address
    ('foo@example.com', 'foo@example.com', 'encode'),
    # Two addresses
    ('foo@example.com, bar@example.com', 'foo@example.com, bar@example.com', 'encode'),
    # Mixed addresses
    ('foo@example.com, Foo Bar <bar@example.com>', 'foo@example.com, Foo Bar <bar@example.com>', 'encode'),
    # Mixed Unicode
    ('foo@example.com, Foo Bar <bar@example.com>, Fôo Baz <baz@example.com>',
     'foo@example.com, Foo Bar <bar@example.com>, \n =?utf-8?q?F=C3=B4o_Baz?= <baz@example.com>', 'encode'),
    ('foo@example.com, Foo Bar <bar@example.com>, Fôo Baz <baz@example.com>, "Quux, Foo" <quux@example.com>',
     ('foo@example.com, Foo Bar <bar@example.com>, \n'
      ' =?utf-8?q?F=C3=B4o_Baz?= <baz@example.com>, "Quux, Foo" <quux@example.com>'), 'encode'),
    ('01234567890123456789012345678901234567890123456789012345678901@example.org, ä <foo@example.org>',
     ('01234567890123456789012345678901234567890123456789012345678901@example.org, \n'
      ' =?utf-8?q?=C3=A4?= <foo@example.org>'), 'encode'),
    # Test for https://github.com/python/cpython/issues/100900
    ('foo@example.com, Foo Bar <bar@example.com>, Fôo Baz <baz@example.com>, "Quûx, Foo" <quux@example.com>',
     ('foo@example.com, Foo Bar <bar@example.com>, \n'
      ' =?utf-8?q?F=C3=B4o_Baz?= <baz@example.com>, \n =?utf-8?q?Qu=C3=BBx=2C_Foo?= <quux@example.com>'), 'encode'),
    # Test preserve
    ('foo@example.com, Foo Bar <bar@example.com>, Fôo Baz <baz@example.com>, "Quûx, Foo" <quux@example.com>',
     'foo@example.com, Foo Bar <bar@example.com>, Fôo Baz <baz@example.com>, \n "Quûx, Foo" <quux@example.com>',
     'preserve'),
    # Test decode
    ('foo@example.com, Foo Bar <bar@example.com>, =?utf-8?q?Qu=C3=BBx=2C_Foo?= <quux@example.com>',
     'foo@example.com, Foo Bar <bar@example.com>, \n "Quûx, Foo" <quux@example.com>',
     'decode'),
    # Test short message-id
    ('Message-ID: <20240319-short-message-id@example.com>', '<20240319-short-message-id@example.com>', 'encode'),
    # Test long message-id
    ('Message-ID: <20240319-very-long-message-id-that-spans-multiple-lines-for-sure-because-longer-than-75-characters-abcde123456@longdomain.example.com>',
     '<20240319-very-long-message-id-that-spans-multiple-lines-for-sure-because-longer-than-75-characters-abcde123456@longdomain.example.com>',
     'encode'),
])
def test_header_wrapping(sampledir: str, hval: str, verify: str, tr: Literal['encode', 'decode', 'preserve']) -> None:
    if ':' in hval:
        chunks = hval.split(':', maxsplit=1)
        hname = chunks[0].strip()
        hval = chunks[1].strip()
    else:
        hname = 'To' if '@' in hval else "X-Header"
    wrapped = b4.LoreMessage.wrap_header((hname, hval), transform=tr)
    assert wrapped.decode() == f'{hname}: {verify}'
    _wname, wval = wrapped.split(b':', maxsplit=1)
    if tr != 'decode':
        cval = b4.LoreMessage.clean_header(wval.decode())
        assert cval == hval


@pytest.mark.parametrize('pairs,verify,clean', [
    ([('', 'foo@example.com'), ('Foo Bar', 'bar@example.com')],
     'foo@example.com, Foo Bar <bar@example.com>', True),
    ([('', 'foo@example.com'), ('Foo, Bar', 'bar@example.com')],
     'foo@example.com, "Foo, Bar" <bar@example.com>', True),
    ([('', 'foo@example.com'), ('Fôo, Bar', 'bar@example.com')],
     'foo@example.com, "Fôo, Bar" <bar@example.com>', True),
    ([('', 'foo@example.com'), ('=?utf-8?q?Qu=C3=BBx_Foo?=', 'quux@example.com')],
     'foo@example.com, Quûx Foo <quux@example.com>', True),
    ([('', 'foo@example.com'), ('=?utf-8?q?Qu=C3=BBx=2C_Foo?=', 'quux@example.com')],
     'foo@example.com, "Quûx, Foo" <quux@example.com>', True),
    ([('', 'foo@example.com'), ('=?utf-8?q?Qu=C3=BBx=2C_Foo?=', 'quux@example.com')],
     'foo@example.com, =?utf-8?q?Qu=C3=BBx=2C_Foo?= <quux@example.com>', False),
    # Pre-quoted display name with special chars must not be double-quoted
    ([('', 'foo@example.com'), ('"Example.org Tools"', 'tools@example.org')],
     'foo@example.com, "Example.org Tools" <tools@example.org>', True),
    ([('', 'foo@example.com'), ('"Doe, Jane"', 'jane@example.com')],
     'foo@example.com, "Doe, Jane" <jane@example.com>', True),
    # Unquoted name with internal quotes
    ([('', 'foo@example.com'), ('Jane "JD" Doe', 'jd@example.com')],
     'foo@example.com, "Jane \\"JD\\" Doe" <jd@example.com>', True),
    # Name starting with quote but not fully quoted
    ([('', 'foo@example.com'), ('"JD" Doe', 'jd@example.com')],
     'foo@example.com, "\\"JD\\" Doe" <jd@example.com>', True),
    # Pre-quoted name with internal quotes
    ([('', 'foo@example.com'), ('"Jane "JD" Doe"', 'jd@example.com')],
     'foo@example.com, "Jane \\"JD\\" Doe" <jd@example.com>', True),
])
def test_format_addrs(pairs: List[Tuple[str, str]], verify: str, clean: bool) -> None:
    formatted = b4.format_addrs(pairs, clean)
    assert formatted == verify


@pytest.mark.parametrize('intrange,upper,expected', [
    ('1-3', 5, [1, 2, 3]),
    ('-1', 5, [5]),
    ('1,3-5', 5, [1, 3, 4, 5]),
    ('1', 5, [1]),
    ('3', 5, [3]),
    ('5', 5, [5]),
    ('1,3,4-', 6, [1, 3, 4, 5, 6]),
    ('1-3,5,-1', 7, [1, 2, 3, 5, 7]),
    ('-7', 5, []),
    ('1-8', 3, [1, 2, 3]),
])
def test_parse_int_range(intrange: str, upper: int, expected: List[int]) -> None:
    assert list(b4.parse_int_range(intrange, upper)) == expected


@pytest.mark.parametrize('body_link,extra_link,expect_count', [
    # Exact same URL — should dedup to one
    ('https://patch.msgid.link/20240101-test-v1-1-abc123@example.com',
     'https://patch.msgid.link/20240101-test-v1-1-abc123@example.com', 1),
    # Same URL, different case — should still dedup
    ('https://patch.msgid.link/20240101-TEST-V1-1-ABC123@example.com',
     'https://patch.msgid.link/20240101-test-v1-1-abc123@example.com', 1),
    # Different domains, same message-id — should dedup to one
    ('https://lore.kernel.org/r/20240101-test-v1-1-abc123@example.com',
     'https://patch.msgid.link/20240101-test-v1-1-abc123@example.com', 1),
    # URL-encoded message-id — should match decoded form
    ('https://lore.kernel.org/r/20240101-test-v1-1-abc123%40example.com',
     'https://patch.msgid.link/20240101-test-v1-1-abc123@example.com', 1),
    # Different message-ids — both should survive
    ('https://lore.kernel.org/r/20240101-foo-v1-1-aaa@example.com',
     'https://patch.msgid.link/20240101-bar-v1-1-bbb@example.com', 2),
])
def test_link_trailer_dedup(body_link: str, extra_link: str, expect_count: int) -> None:
    """Link: trailers already in the body should not be duplicated by extras."""
    raw = (
        f'From: Test Author <test@example.com>\n'
        f'Subject: [PATCH] test link dedup\n'
        f'Date: Mon, 1 Jan 2024 00:00:00 +0000\n'
        f'Message-Id: <20240101-test-v1-1-abc123@example.com>\n'
        f'\n'
        f'Commit body here.\n'
        f'\n'
        f'Signed-off-by: Test Author <test@example.com>\n'
        f'Link: {body_link}\n'
    )
    msg = email.message_from_string(raw, policy=email.policy.EmailPolicy(utf8=True))
    lmsg = b4.LoreMessage(msg)
    extra = b4.LoreTrailer(name='Link', value=extra_link)
    lmsg.fix_trailers(extras=[extra])
    # Count Link: trailers in the result
    _, _, trailers, _, _ = b4.LoreMessage.get_body_parts(lmsg.body)
    link_trailers = [t for t in trailers if t.lname == 'link']
    assert len(link_trailers) == expect_count


class TestTakeFlow:
    """Simulate the 'take' flow using the actual code path: build email
    messages (as if fetched from lore), feed through LoreMailbox →
    LoreSeries → get_am_ready(addlink=True) → git am.

    No network access — messages are constructed in-memory.
    """

    @staticmethod
    def _make_patch_msg(msgid: str, subject: str, body: str,
                        diff: str, counter: int = 1, expected: int = 1,
                        in_reply_to: Optional[str] = None) -> email.message.EmailMessage:
        """Build a realistic patch email like what lore returns.

        The *body* should contain the full commit message including
        trailers (Signed-off-by, Link, etc.) — just like a real patch
        email from a mailing list.
        """
        if expected > 1:
            prefix = f'[PATCH {counter}/{expected}]'
        else:
            prefix = '[PATCH]'
        raw = (
            f'From: Test Author <test@example.com>\n'
            f'Subject: {prefix} {subject}\n'
            f'Date: Mon, 1 Jan 2024 00:00:00 +0000\n'
            f'Message-Id: <{msgid}>\n'
        )
        if in_reply_to:
            raw += f'In-Reply-To: <{in_reply_to}>\n'
            raw += f'References: <{in_reply_to}>\n'
        raw += (
            f'\n'
            f'{body}\n'
            f'---\n'
            f'{diff}\n'
        )
        return email.message_from_string(
            raw, policy=email.policy.EmailPolicy(utf8=True))

    @staticmethod
    def _make_reply_msg(msgid: str, in_reply_to: str,
                        from_name: str, from_email: str,
                        trailer_lines: List[str]) -> email.message.EmailMessage:
        """Build a followup reply with trailers."""
        trailers = '\n'.join(trailer_lines)
        raw = (
            f'From: {from_name} <{from_email}>\n'
            f'Subject: Re: [PATCH] test\n'
            f'Date: Mon, 1 Jan 2024 01:00:00 +0000\n'
            f'Message-Id: <{msgid}>\n'
            f'In-Reply-To: <{in_reply_to}>\n'
            f'References: <{in_reply_to}>\n'
            f'\n'
            f'> Some quoted text\n'
            f'\n'
            f'{trailers}\n'
        )
        return email.message_from_string(
            raw, policy=email.policy.EmailPolicy(utf8=True))

    def test_link_dedup_with_followups(self, gitdir: str) -> None:
        """Patch already has Link: in body, get_am_ready(addlink=True)
        should not duplicate it.  Followup trailers should be added."""
        patch_msgid = '20240101-widget-v1-1-abc123@example.com'
        link_url = f'https://patch.msgid.link/{patch_msgid}'

        patch_msg = self._make_patch_msg(
            msgid=patch_msgid,
            subject='Add widget support',
            body=(
                'This adds a fancy widget.\n'
                '\n'
                'Signed-off-by: Test Author <test@example.com>\n'
                f'Link: {link_url}\n'
            ),
            diff=(
                ' file1.txt | 1 +\n'
                ' 1 file changed, 1 insertion(+)\n'
                '\n'
                'diff --git a/file1.txt b/file1.txt\n'
                'index b352682..6713e9f 100644\n'
                '--- a/file1.txt\n'
                '+++ b/file1.txt\n'
                '@@ -1,3 +1,4 @@\n'
                ' This is file 1.\n'
                ' It has a single line.\n'
                ' This is a second line I added.\n'
                '+widget\n'
            ),
        )

        reply_msg = self._make_reply_msg(
            msgid='reply-1@example.com',
            in_reply_to=patch_msgid,
            from_name='Reviewer One',
            from_email='reviewer@example.com',
            trailer_lines=[
                'Reviewed-by: Reviewer One <reviewer@example.com>',
            ],
        )

        reply_msg2 = self._make_reply_msg(
            msgid='reply-2@example.com',
            in_reply_to=patch_msgid,
            from_name='Acker Two',
            from_email='acker@example.com',
            trailer_lines=[
                'Acked-by: Acker Two <acker@example.com>',
            ],
        )

        # Feed through LoreMailbox → LoreSeries (actual take code path)
        lmbx = b4.LoreMailbox()
        for msg in [patch_msg, reply_msg, reply_msg2]:
            lmbx.add_message(msg)

        lser = lmbx.get_series()
        assert lser is not None

        am_msgs = lser.get_am_ready(addlink=True)
        assert len(am_msgs) == 1

        # Apply to master via git am
        ifh = io.BytesIO()
        b4.save_git_am_mbox(am_msgs, ifh)
        ecode, out = b4.git_run_command(
            gitdir, ['am'], stdin=ifh.getvalue())
        assert ecode == 0, f'git am failed: {out}'

        ecode, result = b4.git_run_command(
            gitdir, ['log', '-1', '--format=%B'])
        assert ecode == 0

        # Exactly one Link: trailer, not two
        assert result.count(f'Link: {link_url}') == 1, \
            f'Duplicate Link: found:\n{result}'
        # Followup trailers applied
        assert 'Reviewed-by: Reviewer One <reviewer@example.com>' in result
        assert 'Acked-by: Acker Two <acker@example.com>' in result

    def test_link_added_when_not_present(self, gitdir: str) -> None:
        """Patch without Link: should get one added by addlink=True."""
        patch_msgid = '20240101-cursor-v1-1-def456@example.com'

        patch_msg = self._make_patch_msg(
            msgid=patch_msgid,
            subject='Fix cursor rendering',
            body=(
                'This fixes a cursor bug.\n'
                '\n'
                'Signed-off-by: Test Author <test@example.com>\n'
            ),
            diff=(
                ' file1.txt | 1 +\n'
                ' 1 file changed, 1 insertion(+)\n'
                '\n'
                'diff --git a/file1.txt b/file1.txt\n'
                'index b352682..e147dad 100644\n'
                '--- a/file1.txt\n'
                '+++ b/file1.txt\n'
                '@@ -1,3 +1,4 @@\n'
                ' This is file 1.\n'
                ' It has a single line.\n'
                ' This is a second line I added.\n'
                '+cursor fix\n'
            ),
        )

        lmbx = b4.LoreMailbox()
        lmbx.add_message(patch_msg)
        lser = lmbx.get_series()
        assert lser is not None

        am_msgs = lser.get_am_ready(addlink=True)
        assert len(am_msgs) == 1

        ifh = io.BytesIO()
        b4.save_git_am_mbox(am_msgs, ifh)
        ecode, out = b4.git_run_command(
            gitdir, ['am'], stdin=ifh.getvalue())
        assert ecode == 0, f'git am failed: {out}'

        ecode, result = b4.git_run_command(
            gitdir, ['log', '-1', '--format=%B'])
        assert ecode == 0

        expected_link = f'https://patch.msgid.link/{patch_msgid}'
        assert f'Link: {expected_link}' in result
        assert result.count('Link:') == 1

    def test_followup_trailers_without_addlink(self, gitdir: str) -> None:
        """Followups should be applied even with addlink=False."""
        patch_msgid = '20240101-verifier-v1-1-789abc@example.com'

        patch_msg = self._make_patch_msg(
            msgid=patch_msgid,
            subject='Refactor verifier',
            body=(
                'Clean up the verifier logic.\n'
                '\n'
                'Signed-off-by: Test Author <test@example.com>\n'
            ),
            diff=(
                ' file1.txt | 1 +\n'
                ' 1 file changed, 1 insertion(+)\n'
                '\n'
                'diff --git a/file1.txt b/file1.txt\n'
                'index b352682..6a8b771 100644\n'
                '--- a/file1.txt\n'
                '+++ b/file1.txt\n'
                '@@ -1,3 +1,4 @@\n'
                ' This is file 1.\n'
                ' It has a single line.\n'
                ' This is a second line I added.\n'
                '+verifier\n'
            ),
        )

        reply_msg = self._make_reply_msg(
            msgid='reply-v-1@example.com',
            in_reply_to=patch_msgid,
            from_name='Alice Author',
            from_email='alice@example.com',
            trailer_lines=[
                'Reviewed-by: Alice Author <alice@example.com>',
                'Tested-by: Alice Author <alice@example.com>',
            ],
        )

        lmbx = b4.LoreMailbox()
        for msg in [patch_msg, reply_msg]:
            lmbx.add_message(msg)
        lser = lmbx.get_series()
        assert lser is not None

        am_msgs = lser.get_am_ready(addlink=False)
        assert len(am_msgs) == 1

        ifh = io.BytesIO()
        b4.save_git_am_mbox(am_msgs, ifh)
        ecode, out = b4.git_run_command(
            gitdir, ['am'], stdin=ifh.getvalue())
        assert ecode == 0, f'git am failed: {out}'

        ecode, result = b4.git_run_command(
            gitdir, ['log', '-1', '--format=%B'])
        assert ecode == 0

        assert 'Reviewed-by: Alice Author <alice@example.com>' in result
        assert 'Tested-by: Alice Author <alice@example.com>' in result
        assert 'Link:' not in result

    def test_different_link_domains_same_msgid_deduped(self, gitdir: str) -> None:
        """If the patch body has a lore.kernel.org Link: and addlink
        generates a patch.msgid.link one for the same message-id,
        only the original should survive (dedup by message-id)."""
        patch_msgid = '20240101-drm-v1-1-aabbcc@example.com'
        lore_link = f'https://lore.kernel.org/r/{patch_msgid}'

        patch_msg = self._make_patch_msg(
            msgid=patch_msgid,
            subject='Fix DRM issue',
            body=(
                'Fix the DRM subsystem.\n'
                '\n'
                'Signed-off-by: Test Author <test@example.com>\n'
                f'Link: {lore_link}\n'
            ),
            diff=(
                ' file1.txt | 1 +\n'
                ' 1 file changed, 1 insertion(+)\n'
                '\n'
                'diff --git a/file1.txt b/file1.txt\n'
                'index b352682..4a2161b 100644\n'
                '--- a/file1.txt\n'
                '+++ b/file1.txt\n'
                '@@ -1,3 +1,4 @@\n'
                ' This is file 1.\n'
                ' It has a single line.\n'
                ' This is a second line I added.\n'
                '+drm fix\n'
            ),
        )

        lmbx = b4.LoreMailbox()
        lmbx.add_message(patch_msg)
        lser = lmbx.get_series()
        assert lser is not None

        am_msgs = lser.get_am_ready(addlink=True)
        assert len(am_msgs) == 1

        ifh = io.BytesIO()
        b4.save_git_am_mbox(am_msgs, ifh)
        ecode, out = b4.git_run_command(
            gitdir, ['am'], stdin=ifh.getvalue())
        assert ecode == 0, f'git am failed: {out}'

        ecode, result = b4.git_run_command(
            gitdir, ['log', '-1', '--format=%B'])
        assert ecode == 0

        # Same message-id in both URLs, so deduped to one Link:
        assert lore_link in result
        assert result.count('Link:') == 1


@pytest.mark.parametrize('subject,extras,expected', [
    ('[PATCH] This is a patch', None, '[PATCH] This is a patch'),
    ('[PATCH v3] This is a patch', None, '[PATCH v3] This is a patch'),
    ('[PATCH RFC v3] This is a patch', None, '[PATCH RFC v3] This is a patch'),
    ('[RFC PATCH v3 1/3] This is a patch', None, '[RFC PATCH v3 1/3] This is a patch'),
    ('[RESEND PATCH v3 1/3] This is a patch', None, '[RESEND PATCH v3 1/3] This is a patch'),
    ('[PATCH RFC v3 2/3] This is a patch', ['RFC'], '[PATCH RFC v3 2/3] This is a patch'),
    ('[PATCH RFC v3 3/12] This is a patch', None, '[PATCH RFC v3 03/12] This is a patch'),
    ('[PATCH RFC v3] This is a [patch]', ['RFC'], '[PATCH RFC v3] This is a [patch]'),
    ('[PATCH RFC v3 2/3] This is a patch', ['netdev', 'bpf'], '[PATCH RFC netdev bpf v3 2/3] This is a patch'),
])
def test_lore_subject_prefixes(subject: str, extras: Optional[List[str]], expected: str) -> None:
    lsubj = b4.LoreSubject(subject)
    assert lsubj.get_rebuilt_subject(eprefixes=extras) == expected



import argparse
import json
import os
import pytest
import b4
import b4.mbox

from typing import Any, Dict, Optional, Tuple
from unittest.mock import patch


class TestAmConflictError:
    """Tests for the AmConflictError exception class."""

    def test_stores_worktree_path_and_output(self) -> None:
        exc = b4.AmConflictError('/tmp/worktree', 'patch failed to apply')
        assert exc.worktree_path == '/tmp/worktree'
        assert exc.output == 'patch failed to apply'

    def test_inherits_from_runtime_error(self) -> None:
        exc = b4.AmConflictError('/tmp/wt', 'conflict')
        assert isinstance(exc, RuntimeError)

    def test_catchable_as_runtime_error(self) -> None:
        with pytest.raises(RuntimeError):
            raise b4.AmConflictError('/tmp/wt', 'conflict')

    def test_str_is_output(self) -> None:
        exc = b4.AmConflictError('/tmp/wt', 'the error output')
        assert str(exc) == 'the error output'

    def test_empty_output(self) -> None:
        exc = b4.AmConflictError('/tmp/wt', '')
        assert exc.output == ''
        assert exc.worktree_path == '/tmp/wt'


class TestRewriteFetchHeadOrigin:
    """Tests for the _rewrite_fetch_head_origin helper."""

    def test_rewrites_worktree_path(self, gitdir: str) -> None:
        fh_path = os.path.join(gitdir, '.git', 'FETCH_HEAD')
        with open(fh_path, 'w') as fh:
            fh.write("abc123\t\tnot-for-merge\tbranch 'master' of /tmp/b4-worktree\n")

        b4._rewrite_fetch_head_origin(gitdir, '/tmp/b4-worktree',
                                      'https://lore.kernel.org/r/test@msg')

        with open(fh_path, 'r') as fh:
            contents = fh.read()
        assert '/tmp/b4-worktree' not in contents
        assert 'patches from https://lore.kernel.org/r/test@msg' in contents

    def test_noop_when_old_origin_absent(self, gitdir: str) -> None:
        fh_path = os.path.join(gitdir, '.git', 'FETCH_HEAD')
        original = "abc123\t\tnot-for-merge\tbranch 'master' of /some/other/path\n"
        with open(fh_path, 'w') as fh:
            fh.write(original)

        b4._rewrite_fetch_head_origin(gitdir, '/tmp/nonexistent',
                                      'https://example.com')

        with open(fh_path, 'r') as fh:
            contents = fh.read()
        assert contents == original

    def test_rewrites_multiple_occurrences(self, gitdir: str) -> None:
        fh_path = os.path.join(gitdir, '.git', 'FETCH_HEAD')
        with open(fh_path, 'w') as fh:
            fh.write("aaa\t\tnot-for-merge\tbranch 'master' of /tmp/wt\n"
                     "bbb\t\tnot-for-merge\tbranch 'master' of /tmp/wt\n")

        b4._rewrite_fetch_head_origin(gitdir, '/tmp/wt', 'https://example.com')

        with open(fh_path, 'r') as fh:
            contents = fh.read()
        assert contents.count('patches from https://example.com') == 2
        assert '/tmp/wt' not in contents


def _build_clean_patches(gitdir: str) -> Tuple[bytes, str]:
    """Create 2 patches on a temp branch, return (mbox_bytes, base_commit).

    The patches are based on the current HEAD so they apply cleanly
    when at_base=base_commit.
    """
    ecode, base = b4.git_run_command(gitdir, ['rev-parse', 'HEAD'])
    base = base.strip()

    # Create patches on a detached temp branch
    b4.git_run_command(gitdir, ['checkout', '-b', 'clean-patches'])
    with open(os.path.join(gitdir, 'file1.txt'), 'a') as fh:
        fh.write('Added by clean patch 1.\n')
    b4.git_run_command(gitdir, ['add', 'file1.txt'])
    b4.git_run_command(gitdir, ['commit', '-m', 'Clean patch 1'])

    with open(os.path.join(gitdir, 'file1.txt'), 'a') as fh:
        fh.write('Added by clean patch 2.\n')
    b4.git_run_command(gitdir, ['add', 'file1.txt'])
    b4.git_run_command(gitdir, ['commit', '-m', 'Clean patch 2'])

    ecode, mbox = b4.git_run_command(gitdir, ['format-patch', '-2', '--stdout'])
    assert ecode == 0

    # Return to master
    b4.git_run_command(gitdir, ['checkout', 'master'])
    b4.git_run_command(gitdir, ['branch', '-D', 'clean-patches'])

    return mbox.encode(), base


def _build_conflicting_patches(gitdir: str) -> Tuple[bytes, str]:
    """Create a patch that will conflict with a change on master.

    Both the patch and master rewrite all of file1.txt differently,
    so three-way merge detects a conflict.

    Returns (mbox_bytes, base_commit_before_master_modification).
    """
    ecode, base = b4.git_run_command(gitdir, ['rev-parse', 'HEAD'])
    base = base.strip()

    # Create patch on a temp branch (from original HEAD)
    b4.git_run_command(gitdir, ['checkout', '-b', 'conflict-patch'])
    with open(os.path.join(gitdir, 'file1.txt'), 'w') as fh:
        fh.write('PATCH version of file 1.\n'
                 'Rewritten entirely by the patch.\n')
    b4.git_run_command(gitdir, ['add', 'file1.txt'])
    b4.git_run_command(gitdir, ['commit', '-m', 'Rewrite file1 (patch side)'])

    ecode, mbox = b4.git_run_command(gitdir, ['format-patch', '-1', '--stdout'])
    assert ecode == 0

    # Make a conflicting change on master (same lines, different content)
    b4.git_run_command(gitdir, ['checkout', 'master'])
    b4.git_run_command(gitdir, ['branch', '-D', 'conflict-patch'])
    with open(os.path.join(gitdir, 'file1.txt'), 'w') as fh:
        fh.write('MASTER version of file 1.\n'
                 'Also rewritten, but differently.\n')
    b4.git_run_command(gitdir, ['add', 'file1.txt'])
    b4.git_run_command(gitdir, ['commit', '-m', 'Rewrite file1 (master side)'])

    return mbox.encode(), base


class TestGitFetchAmIntoRepo:
    """Integration tests for git_fetch_am_into_repo with three-way merge."""

    def test_clean_apply_with_three_way(self, gitdir: str) -> None:
        """Patches apply cleanly with -3 flag, worktree is cleaned up."""
        ambytes, base = _build_clean_patches(gitdir)
        common_dir = b4.git_get_common_dir(gitdir)
        assert common_dir is not None
        gwt = os.path.join(common_dir, 'b4-shazam-worktree')

        b4.git_fetch_am_into_repo(gitdir, ambytes, at_base=base,
                                  am_flags=['-3'])

        # Worktree should be cleaned up after success
        assert not os.path.exists(gwt)

        # FETCH_HEAD should exist
        fh_path = os.path.join(gitdir, '.git', 'FETCH_HEAD')
        assert os.path.exists(fh_path)

    def test_clean_apply_with_origin_rewrites_fetch_head(self, gitdir: str) -> None:
        """When origin is provided, FETCH_HEAD is rewritten to show it."""
        ambytes, base = _build_clean_patches(gitdir)
        origin = 'https://lore.kernel.org/r/test@example.com'

        b4.git_fetch_am_into_repo(gitdir, ambytes, at_base=base,
                                  origin=origin, am_flags=['-3'])

        fh_path = os.path.join(gitdir, '.git', 'FETCH_HEAD')
        with open(fh_path, 'r') as fh:
            contents = fh.read()
        assert f'patches from {origin}' in contents

    def test_conflict_raises_am_conflict_error(self, gitdir: str) -> None:
        """AmConflictError is raised when patches conflict."""
        ambytes, _base = _build_conflicting_patches(gitdir)

        try:
            with pytest.raises(b4.AmConflictError) as exc_info:
                b4.git_fetch_am_into_repo(gitdir, ambytes, at_base='HEAD',
                                          am_flags=['-3'])

            assert exc_info.value.worktree_path != ''
            assert exc_info.value.output != ''
        finally:
            # Clean up worktree
            common_dir = b4.git_get_common_dir(gitdir)
            if common_dir:
                gwt = os.path.join(common_dir, 'b4-shazam-worktree')
                if os.path.exists(gwt):
                    b4.git_run_command(gitdir, ['worktree', 'remove', '--force', gwt])

    def test_conflict_preserves_worktree(self, gitdir: str) -> None:
        """On conflict, the worktree is preserved for user resolution."""
        ambytes, _base = _build_conflicting_patches(gitdir)

        try:
            with pytest.raises(b4.AmConflictError) as exc_info:
                b4.git_fetch_am_into_repo(gitdir, ambytes, at_base='HEAD',
                                          am_flags=['-3'])

            wt_path = exc_info.value.worktree_path
            # Worktree must still exist for user to resolve
            assert os.path.isdir(wt_path)
            # rebase-apply should be present (am still in progress)
            ecode, wt_gitdir = b4.git_run_command(
                wt_path, ['rev-parse', '--git-dir'],
                logstderr=True, rundir=wt_path)
            assert ecode == 0
            rebase_apply = os.path.join(wt_gitdir.strip(), 'rebase-apply')
            assert os.path.isdir(rebase_apply)
        finally:
            # Clean up worktree
            common_dir = b4.git_get_common_dir(gitdir)
            if common_dir:
                gwt = os.path.join(common_dir, 'b4-shazam-worktree')
                if os.path.exists(gwt):
                    b4.git_run_command(gitdir, ['worktree', 'remove', '--force', gwt])

    def test_clean_apply_without_three_way(self, gitdir: str) -> None:
        """Patches also apply cleanly without -3 (baseline)."""
        ambytes, base = _build_clean_patches(gitdir)

        b4.git_fetch_am_into_repo(gitdir, ambytes, at_base=base,
                                  am_flags=[])

        fh_path = os.path.join(gitdir, '.git', 'FETCH_HEAD')
        assert os.path.exists(fh_path)

    def test_check_only_with_three_way(self, gitdir: str) -> None:
        """check_only mode returns early without fetching, even with -3."""
        ambytes, base = _build_clean_patches(gitdir)

        # Remove any existing FETCH_HEAD
        fh_path = os.path.join(gitdir, '.git', 'FETCH_HEAD')
        if os.path.exists(fh_path):
            os.unlink(fh_path)

        b4.git_fetch_am_into_repo(gitdir, ambytes, at_base=base,
                                  check_only=True, am_flags=['-3'])

        # check_only should not fetch (no FETCH_HEAD created)
        assert not os.path.exists(fh_path)

        # Worktree should still be cleaned up
        common_dir = b4.git_get_common_dir(gitdir)
        assert common_dir is not None
        gwt = os.path.join(common_dir, 'b4-shazam-worktree')
        assert not os.path.exists(gwt)


class TestSuspendToShellCwd:
    """Test that _suspend_to_shell passes cwd to subprocess.run."""

    @patch('b4.review_tui._common.subprocess.run')
    def test_cwd_passed_through(self, mock_run: Any,
                                monkeypatch: pytest.MonkeyPatch) -> None:
        from b4.review_tui._common import _suspend_to_shell
        # Use a shell name that is neither bash nor zsh so we hit
        # the simple else branch (no tempfile/rcfile logic).
        monkeypatch.setenv('SHELL', '/tmp/fakeshell')

        _suspend_to_shell(cwd='/tmp/test-worktree')

        mock_run.assert_called_once()
        _args, kwargs = mock_run.call_args
        assert kwargs.get('cwd') == '/tmp/test-worktree'

    @patch('b4.review_tui._common.subprocess.run')
    def test_cwd_none_by_default(self, mock_run: Any,
                                 monkeypatch: pytest.MonkeyPatch) -> None:
        from b4.review_tui._common import _suspend_to_shell
        monkeypatch.setenv('SHELL', '/tmp/fakeshell')

        _suspend_to_shell()

        mock_run.assert_called_once()
        _args, kwargs = mock_run.call_args
        assert kwargs.get('cwd') is None

    @patch('b4.review_tui._common.subprocess.run')
    def test_hint_appears_in_env(self, mock_run: Any,
                                 monkeypatch: pytest.MonkeyPatch) -> None:
        from b4.review_tui._common import _suspend_to_shell
        monkeypatch.setenv('SHELL', '/tmp/fakeshell')

        _suspend_to_shell(hint='b4 conflict', cwd='/tmp/wt')

        mock_run.assert_called_once()
        _args, kwargs = mock_run.call_args
        assert kwargs['env']['B4_REVIEW'] == 'b4 conflict'


class TestConflictResolutionFlow:
    """Integration tests for the conflict resolution workflow.

    These exercise the same git operations that the TUI conflict
    handlers use (sparse-checkout disable, rebase-apply detection,
    fetch from worktree), without requiring the TUI itself.
    """

    def test_worktree_resolve_and_fetch(self, gitdir: str) -> None:
        """Full worktree flow: conflict -> resolve -> fetch succeeds."""
        ambytes, _base = _build_conflicting_patches(gitdir)

        with pytest.raises(b4.AmConflictError) as exc_info:
            b4.git_fetch_am_into_repo(gitdir, ambytes, at_base='HEAD',
                                      am_flags=['-3'])

        wt = exc_info.value.worktree_path

        # --- same steps the TUI handler takes ---
        # 1. Disable sparse checkout so files are visible
        b4.git_run_command(wt, ['sparse-checkout', 'disable'],
                           logstderr=True, rundir=wt)
        assert os.path.exists(os.path.join(wt, 'file1.txt'))

        # 2. Simulate user resolving: accept theirs and continue
        b4.git_run_command(wt, ['checkout', '--theirs', '.'],
                           logstderr=True, rundir=wt)
        b4.git_run_command(wt, ['add', '-A'],
                           logstderr=True, rundir=wt)
        ecode, _out = b4.git_run_command(wt, ['am', '--continue'],
                                         logstderr=True, rundir=wt)
        assert ecode == 0

        # 3. Verify rebase-apply is gone (am completed)
        ecode, wt_gitdir = b4.git_run_command(
            wt, ['rev-parse', '--git-dir'],
            logstderr=True, rundir=wt)
        assert ecode == 0
        rebase_apply = os.path.join(wt_gitdir.strip(), 'rebase-apply')
        assert not os.path.isdir(rebase_apply)

        # 4. Fetch result into main repo
        ecode, _out = b4.git_run_command(gitdir, ['fetch', wt], logstderr=True)
        assert ecode == 0
        fh_path = os.path.join(gitdir, '.git', 'FETCH_HEAD')
        assert os.path.exists(fh_path)

        # 5. Clean up worktree
        b4.git_run_command(gitdir, ['worktree', 'remove', '--force', wt])
        assert not os.path.exists(wt)

    def test_worktree_unresolved_detected(self, gitdir: str) -> None:
        """When user doesn't resolve, rebase-apply is still present."""
        ambytes, _base = _build_conflicting_patches(gitdir)

        with pytest.raises(b4.AmConflictError) as exc_info:
            b4.git_fetch_am_into_repo(gitdir, ambytes, at_base='HEAD',
                                      am_flags=['-3'])

        wt = exc_info.value.worktree_path

        # User returns from shell without resolving
        ecode, wt_gitdir = b4.git_run_command(
            wt, ['rev-parse', '--git-dir'],
            logstderr=True, rundir=wt)
        assert ecode == 0
        rebase_apply = os.path.join(wt_gitdir.strip(), 'rebase-apply')
        assert os.path.isdir(rebase_apply)

        # Handler cleans up the worktree
        b4.git_run_command(gitdir, ['worktree', 'remove', '--force', wt])
        assert not os.path.exists(wt)

    def test_sparse_checkout_disable_exposes_files(self, gitdir: str) -> None:
        """Disabling sparse checkout makes worktree files visible."""
        ambytes, _base = _build_conflicting_patches(gitdir)

        with pytest.raises(b4.AmConflictError) as exc_info:
            b4.git_fetch_am_into_repo(gitdir, ambytes, at_base='HEAD',
                                      am_flags=['-3'])

        wt = exc_info.value.worktree_path

        # Before: sparse checkout may hide files
        # (the worktree was created with sparse-checkout set to empty)
        b4.git_run_command(wt, ['sparse-checkout', 'disable'],
                           logstderr=True, rundir=wt)

        # All repo files should now be visible
        assert os.path.exists(os.path.join(wt, 'file1.txt'))
        assert os.path.exists(os.path.join(wt, 'file2.txt'))
        assert os.path.exists(os.path.join(wt, 'lipsum.txt'))

        b4.git_run_command(gitdir, ['worktree', 'remove', '--force', wt])

    def test_fetch_head_origin_rewrite_after_resolve(self, gitdir: str) -> None:
        """After resolving and fetching, FETCH_HEAD origin is rewritten."""
        ambytes, _base = _build_conflicting_patches(gitdir)

        with pytest.raises(b4.AmConflictError) as exc_info:
            b4.git_fetch_am_into_repo(gitdir, ambytes, at_base='HEAD',
                                      am_flags=['-3'])

        wt = exc_info.value.worktree_path

        # Resolve and fetch
        b4.git_run_command(wt, ['sparse-checkout', 'disable'],
                           logstderr=True, rundir=wt)
        b4.git_run_command(wt, ['checkout', '--theirs', '.'],
                           logstderr=True, rundir=wt)
        b4.git_run_command(wt, ['add', '-A'],
                           logstderr=True, rundir=wt)
        b4.git_run_command(wt, ['am', '--continue'],
                           logstderr=True, rundir=wt)
        b4.git_run_command(gitdir, ['fetch', wt], logstderr=True)

        # Rewrite FETCH_HEAD (as the TUI handler does)
        origin = 'https://lore.kernel.org/r/test@example.com'
        b4._rewrite_fetch_head_origin(gitdir, wt, origin)

        fh_path = os.path.join(gitdir, '.git', 'FETCH_HEAD')
        with open(fh_path, 'r') as fh:
            contents = fh.read()
        assert wt not in contents
        assert f'patches from {origin}' in contents

        b4.git_run_command(gitdir, ['worktree', 'remove', '--force', wt])


class TestDirectAmConflictFlow:
    """Integration tests for the direct git-am conflict path (_do_take_am).

    This path runs git-am directly on the user's working branch
    (not in a worktree), so resolution happens in-place.
    """

    def test_am_conflict_and_resolution(self, gitdir: str) -> None:
        """Direct git-am -3 conflict, resolved with --continue."""
        ambytes, _base = _build_conflicting_patches(gitdir)

        # Run git-am directly (as _do_take_am does)
        ecode, _out = b4.git_run_command(gitdir, ['am', '-3'],
                                         stdin=ambytes, logstderr=True)
        assert ecode != 0

        # rebase-apply should exist
        rebase_apply = os.path.join(gitdir, '.git', 'rebase-apply')
        assert os.path.isdir(rebase_apply)

        # Resolve: accept theirs and continue
        b4.git_run_command(gitdir, ['checkout', '--theirs', '.'],
                           logstderr=True)
        b4.git_run_command(gitdir, ['add', '-A'], logstderr=True)
        ecode, _out = b4.git_run_command(gitdir, ['am', '--continue'],
                                         logstderr=True)
        assert ecode == 0
        assert not os.path.isdir(rebase_apply)

    def test_am_conflict_unresolved(self, gitdir: str) -> None:
        """Direct git-am -3 conflict, user aborts."""
        ambytes, _base = _build_conflicting_patches(gitdir)

        ecode, _out = b4.git_run_command(gitdir, ['am', '-3'],
                                         stdin=ambytes, logstderr=True)
        assert ecode != 0

        # rebase-apply is present (handler detects this)
        rebase_apply = os.path.join(gitdir, '.git', 'rebase-apply')
        assert os.path.isdir(rebase_apply)

        # Abort (as user would after incomplete resolution)
        b4.git_run_command(gitdir, ['am', '--abort'], logstderr=True)
        assert not os.path.isdir(rebase_apply)


# ---------------------------------------------------------------------------
# Tier 4 — Shazam state machine tests
# ---------------------------------------------------------------------------

def _build_multi_patch_conflict(gitdir: str) -> Tuple[bytes, str]:
    """Create a 3-patch mbox where patches 1-2 are clean but patch 3 conflicts.

    Patches 1-2 modify file2.txt and lipsum.txt (no conflict with master).
    Patch 3 rewrites file1.txt (conflicts with master's rewrite).

    Returns (mbox_bytes, original_base_commit).
    """
    ecode, base = b4.git_run_command(gitdir, ['rev-parse', 'HEAD'])
    base = base.strip()

    # Create 3 patches on a temp branch
    b4.git_run_command(gitdir, ['checkout', '-b', 'multi-patch'])

    # Patch 1: modify file2.txt (clean)
    with open(os.path.join(gitdir, 'file2.txt'), 'a') as fh:
        fh.write('Added by patch 1.\n')
    b4.git_run_command(gitdir, ['add', 'file2.txt'])
    b4.git_run_command(gitdir, ['commit', '-m', 'Patch 1: modify file2'])

    # Patch 2: modify lipsum.txt (clean)
    with open(os.path.join(gitdir, 'lipsum.txt'), 'a') as fh:
        fh.write('\nExtra paragraph from patch 2.\n')
    b4.git_run_command(gitdir, ['add', 'lipsum.txt'])
    b4.git_run_command(gitdir, ['commit', '-m', 'Patch 2: modify lipsum'])

    # Patch 3: rewrite file1.txt (will conflict with master)
    with open(os.path.join(gitdir, 'file1.txt'), 'w') as fh:
        fh.write('PATCH version of file1.\nRewritten by patch 3.\n')
    b4.git_run_command(gitdir, ['add', 'file1.txt'])
    b4.git_run_command(gitdir, ['commit', '-m', 'Patch 3: rewrite file1'])

    ecode, mbox = b4.git_run_command(gitdir, ['format-patch', '-3', '--stdout'])
    assert ecode == 0

    # Back to master, make a conflicting change to file1.txt only
    b4.git_run_command(gitdir, ['checkout', 'master'])
    b4.git_run_command(gitdir, ['branch', '-D', 'multi-patch'])
    with open(os.path.join(gitdir, 'file1.txt'), 'w') as fh:
        fh.write('MASTER version of file1.\nConflicting rewrite.\n')
    b4.git_run_command(gitdir, ['add', 'file1.txt'])
    b4.git_run_command(gitdir, ['commit', '-m', 'Master: rewrite file1'])

    return mbox.encode(), base


def _make_shazam_state(common_dir: str,
                       state: Optional[Dict[str, Any]] = None) -> Tuple[str, str]:
    """Create shazam state file and patches dir.

    Returns (state_file_path, patches_dir_path).
    """
    state_file = os.path.join(common_dir, 'b4-shazam-state.json')
    patches_dir = os.path.join(common_dir, 'b4-shazam-patches')
    os.makedirs(patches_dir, exist_ok=True)
    if state is None:
        state = {'origin': 'https://example.com', 'merge_flags': '--signoff'}
    with open(state_file, 'w') as fh:
        json.dump(state, fh)
    return state_file, patches_dir


class TestLoadShazamState:
    """Tests for _load_shazam_state."""

    def test_valid_state_loaded(self, gitdir: str) -> None:
        common_dir = b4.git_get_common_dir(gitdir)
        assert common_dir is not None
        state_file, patches_dir = _make_shazam_state(common_dir)
        try:
            _topdir, _cdir, sf, loaded = b4.mbox._load_shazam_state(
                require_state=True)
            assert loaded == {'origin': 'https://example.com',
                              'merge_flags': '--signoff'}
            assert sf == state_file
        finally:
            os.unlink(state_file)
            os.rmdir(patches_dir)

    def test_missing_state_exits(self, gitdir: str) -> None:
        with pytest.raises(SystemExit) as exc_info:
            b4.mbox._load_shazam_state(require_state=True)
        assert exc_info.value.code == 1

    def test_optional_state_returns_none(self, gitdir: str) -> None:
        _topdir, _cdir, _sf, loaded = b4.mbox._load_shazam_state(
            require_state=False)
        assert loaded is None

    def test_missing_patches_dir_exits(self, gitdir: str) -> None:
        common_dir = b4.git_get_common_dir(gitdir)
        assert common_dir is not None
        # Create state file but NOT the patches dir
        state_file = os.path.join(common_dir, 'b4-shazam-state.json')
        with open(state_file, 'w') as fh:
            json.dump({'origin': 'test'}, fh)
        try:
            with pytest.raises(SystemExit) as exc_info:
                b4.mbox._load_shazam_state(require_state=True)
            assert exc_info.value.code == 1
        finally:
            if os.path.exists(state_file):
                os.unlink(state_file)


class TestShazamAbort:
    """Tests for shazam_abort cleanup."""

    def test_cleans_up_all_artifacts(self, gitdir: str) -> None:
        common_dir = b4.git_get_common_dir(gitdir)
        assert common_dir is not None

        state_file, patches_dir = _make_shazam_state(common_dir)
        # Add a fake patch file
        with open(os.path.join(patches_dir, '0000'), 'w') as fh:
            fh.write('patch data')

        cmdargs = argparse.Namespace()
        b4.mbox.shazam_abort(cmdargs)

        assert not os.path.exists(patches_dir)
        assert not os.path.exists(state_file)

    def test_cleans_up_stale_worktree(self, gitdir: str) -> None:
        common_dir = b4.git_get_common_dir(gitdir)
        assert common_dir is not None

        state_file, patches_dir = _make_shazam_state(common_dir)

        # Create a stale worktree
        gwt = os.path.join(common_dir, 'b4-shazam-worktree')
        b4.git_run_command(gitdir, ['worktree', 'add', '--detach', gwt, 'HEAD'])
        assert os.path.isdir(gwt)

        cmdargs = argparse.Namespace()
        b4.mbox.shazam_abort(cmdargs)

        assert not os.path.exists(gwt)
        assert not os.path.exists(patches_dir)
        assert not os.path.exists(state_file)

    def test_noop_when_nothing_to_clean(self, gitdir: str) -> None:
        cmdargs = argparse.Namespace()
        # Should not raise
        b4.mbox.shazam_abort(cmdargs)


class TestStartMergeResolve:
    """Integration tests for _start_merge_resolve.

    This function extracts remaining patches from a failed git-am
    worktree, fetches successfully-applied patches, starts a merge,
    and applies remaining patches one-by-one.
    """

    def test_creates_state_and_patches(self, gitdir: str) -> None:
        """After a multi-patch conflict, state files are created."""
        ambytes, _base = _build_multi_patch_conflict(gitdir)
        common_dir = b4.git_get_common_dir(gitdir)
        assert common_dir is not None

        with pytest.raises(b4.AmConflictError) as exc_info:
            b4.git_fetch_am_into_repo(gitdir, ambytes, at_base='HEAD',
                                      am_flags=['-3'])

        state = {
            'origin': 'https://example.com',
            'merge_template_values': {},
            'merge_template': 'Test merge\n\nConflict resolution test.',
            'merge_flags': '--signoff',
            'no_interactive': True,
        }

        # _start_merge_resolve exits(1) because remaining patch 3 conflicts
        with pytest.raises(SystemExit) as exit_info:
            b4.mbox._start_merge_resolve(
                gitdir, exc_info.value, common_dir, state)
        assert exit_info.value.code == 1

        # State file and patches dir should exist
        state_file = os.path.join(common_dir, 'b4-shazam-state.json')
        patches_dir = os.path.join(common_dir, 'b4-shazam-patches')
        assert os.path.exists(state_file)
        assert os.path.isdir(patches_dir)

        # One remaining patch was extracted (patch 3)
        with open(os.path.join(patches_dir, 'total'), 'r') as fh:
            assert fh.read().strip() == '1'

        # Worktree should be removed
        gwt = os.path.join(common_dir, 'b4-shazam-worktree')
        assert not os.path.exists(gwt)

        # Clean up for fixture teardown
        b4.git_run_command(gitdir, ['merge', '--abort'], logstderr=True)

    def test_full_resolve_continue_flow(self, gitdir: str) -> None:
        """Full flow: conflict -> resolve -> shazam --continue -> merge commit."""
        ambytes, _base = _build_multi_patch_conflict(gitdir)
        common_dir = b4.git_get_common_dir(gitdir)
        assert common_dir is not None

        # Step 1: trigger conflict
        with pytest.raises(b4.AmConflictError) as exc_info:
            b4.git_fetch_am_into_repo(gitdir, ambytes, at_base='HEAD',
                                      am_flags=['-3'])

        state = {
            'origin': 'https://example.com',
            'merge_template_values': {},
            'merge_template': 'Test merge\n\nResolved conflict.',
            'merge_flags': '--signoff',
            'no_interactive': True,
        }

        # Step 2: _start_merge_resolve extracts patches, starts merge,
        # applies remaining patch 3 which conflicts -> exit(1)
        with pytest.raises(SystemExit):
            b4.mbox._start_merge_resolve(
                gitdir, exc_info.value, common_dir, state)

        # Step 3: resolve the conflict (accept any content)
        with open(os.path.join(gitdir, 'file1.txt'), 'w') as fh:
            fh.write('Resolved content for file1.\n')
        b4.git_run_command(gitdir, ['add', 'file1.txt'])

        # Step 4: shazam --continue
        cmdargs = argparse.Namespace()
        # Should complete successfully (no SystemExit)
        b4.mbox.shazam_continue(cmdargs)

        # Step 5: verify merge commit was created
        ecode, _log_out = b4.git_run_command(
            gitdir, ['log', '--oneline', '-1', '--format=%s'])
        assert ecode == 0
        # The commit was made with -F (the merge template content)
        # Just verify a commit exists on top of our branch
        ecode, parents = b4.git_run_command(
            gitdir, ['rev-list', '--parents', '-1', 'HEAD'])
        assert ecode == 0
        # Merge commit has 2 parents
        parent_list = parents.strip().split()
        assert len(parent_list) == 3  # commit_hash parent1 parent2

        # State files should be cleaned up
        state_file = os.path.join(common_dir, 'b4-shazam-state.json')
        patches_dir = os.path.join(common_dir, 'b4-shazam-patches')
        assert not os.path.exists(state_file)
        assert not os.path.exists(patches_dir)

    def test_abort_after_conflict(self, gitdir: str) -> None:
        """After conflict, shazam --abort cleans everything up."""
        ambytes, _base = _build_multi_patch_conflict(gitdir)
        common_dir = b4.git_get_common_dir(gitdir)
        assert common_dir is not None

        with pytest.raises(b4.AmConflictError) as exc_info:
            b4.git_fetch_am_into_repo(gitdir, ambytes, at_base='HEAD',
                                      am_flags=['-3'])

        state = {
            'origin': 'https://example.com',
            'merge_template_values': {},
            'merge_template': 'Test merge',
            'merge_flags': '--signoff',
            'no_interactive': True,
        }

        with pytest.raises(SystemExit):
            b4.mbox._start_merge_resolve(
                gitdir, exc_info.value, common_dir, state)

        # Abort instead of resolving
        cmdargs = argparse.Namespace()
        b4.mbox.shazam_abort(cmdargs)

        # Everything should be cleaned up
        state_file = os.path.join(common_dir, 'b4-shazam-state.json')
        patches_dir = os.path.join(common_dir, 'b4-shazam-patches')
        assert not os.path.exists(state_file)
        assert not os.path.exists(patches_dir)

        # Merge should be aborted (no MERGE_HEAD)
        merge_head = os.path.join(gitdir, '.git', 'MERGE_HEAD')
        assert not os.path.exists(merge_head)



project = 'B4 end-user docs'
copyright = '2019-2022, The Linux Foundation and contributors'
author = 'Kernel.org'
templates_path = ['_templates']
html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
html_logo = '_static/b4-logo.png'
# Don't highlight by default
highlight_language = 'none'
master_doc = 'index'



from b4.review_tui._common import display_width, pad_display


class TestDisplayWidth:
    """Tests for display_width()."""

    def test_ascii(self) -> None:
        assert display_width('hello') == 5

    def test_empty(self) -> None:
        assert display_width('') == 0

    def test_cjk(self) -> None:
        # Each CJK character is 2 columns wide
        assert display_width('戸田晃太') == 8

    def test_mixed(self) -> None:
        # 4 CJK chars (8) + 1 space (1) + 3 ASCII (3) = 12
        assert display_width('戸田晃太 abc') == 12

    def test_fullwidth_latin(self) -> None:
        # U+FF21 FULLWIDTH LATIN CAPITAL LETTER A
        assert display_width('\uff21') == 2

    def test_emoji(self) -> None:
        # Most emoji have east_asian_width 'W'
        assert display_width('\u2605') >= 1  # BLACK STAR


class TestPadDisplay:
    """Tests for pad_display()."""

    def test_ascii_padding(self) -> None:
        result = pad_display('hello', 10)
        assert result == 'hello     '
        assert len(result) == 10

    def test_cjk_padding(self) -> None:
        # '戸田' = 4 display cols, pad to 10 = 6 spaces
        result = pad_display('戸田', 10)
        assert display_width(result) == 10
        assert result == '戸田      '

    def test_no_padding_when_exact(self) -> None:
        result = pad_display('hello', 5)
        assert result == 'hello'

    def test_truncate_when_over(self) -> None:
        result = pad_display('hello world', 5)
        assert result == 'hell\u2026'
        assert display_width(result) == 5

    def test_truncate_long_name(self) -> None:
        result = pad_display('Bastien Curutchet (Schneider Electric)', 30)
        assert display_width(result) == 30
        assert result.endswith('\u2026')

    def test_truncate_cjk(self) -> None:
        # '戸田晃太' = 8 display cols, truncate to 5: '戸田' (4) + ellipsis (1)
        result = pad_display('戸田晃太', 5)
        assert display_width(result) == 5
        assert result.endswith('\u2026')

    def test_mixed_padding(self) -> None:
        # 'K 戸田' = 1 + 1 + 2 + 2 = 6 display cols, pad to 10 = 4 spaces
        result = pad_display('K 戸田', 10)
        assert display_width(result) == 10
